# handle the tal: namespace

tal_ns = 'http://xml.zope.org/namespaces/tal'

from pax.paxtransform import AttributeHandler, CompiledInformationHandler, \
     AttributeCompiler, StopTransform
from pax import XML
from pax.backwards_compatibility import *
import re

def normalize(text):
    return ' '.join(text.split())

def text_or_structure(value):
    return split_spec(value, 'text', 'structure')

def split_spec(value, *names):
    for name in names:
        if value.startswith(name + ' '):
            return name, value[len(name)+1:]
    return names[0], value

def remove_attrs_till(name, element):
    data = getattr(element, tal_handler.compiled_name)
    pname = None
    while data and pname != name:
        pname, pdata = data.pop(0)

class Structure(object):
    def __init__(self, text):
        self.text = text

    def __xml__(self, *a):
        return self.text

class Default(Exception):
    pass

multiple_re = re.compile('(?<!;);(?!;)')

def split_multiple(s):
    """split multiple semicolon-separated values
    two semicolons should be considered an escape"""
    parts = [sub.strip().replace(';;', ';') for sub in multiple_re.split(s)]
    if len(parts) > 1 and not parts[-1].strip():
        del parts[-1] # It ended in a semicolon
    return parts


def handle_define(element, context, value):
    for is_global, name, expr in value:
        result = context.engine.evaluateValue(expr)
        #print 'defining (%s) %s as %s' % (is_global, name, result)
        if is_global:
            context.engine.setGlobal(name, result)
        else:
            context.engine.setLocal(name, result)
    return element

def handle_condition(element, context, value):
    if context.engine.evaluateBoolean(value):
        return element
    raise StopTransform

def handle_repeat(element, context, value):
    remove_attrs_till('repeat', element)
    result = []
    iterator = context.engine.setRepeat(*value)
    while iterator.next():
        result.append(context.tr_engine.transform(element, context))
    raise StopTransform, (result,)

def get_text_or_structure(context, structure, value):
    content = context.engine.evaluateText(value)
    if content is context.engine.getDefault():
        raise Default
    if content and structure:
        content = Structure(content)
    return content

def handle_replace(element, context, value):
    try:
        content = get_text_or_structure(context, *value)
    except Default:
        return element
    raise StopTransform, content

def handle_content(element, context, value):
    try:
        content = get_text_or_structure(context, *value)
    except Default:
        return element
    element.children = [content]
    return element

def handle_attributes(element, context, value):
    for ns, name, expr in value:
        value = context.engine.evaluateText(expr)
        if value is not context.engine.getDefault():
            nsattrs = element.attributes.setdefault(ns, {})
            if value:
                nsattrs[name] = value
            else:
                try:
                    del nsattrs[name]
                except KeyError:
                    pass
    return element

def handle_omit_tag(element, context, value):
    if value and not context.engine.evaluateBoolean(value):
        return element
    return element.children

def tal_start(element, context):
    context.engine.beginScope()
    context.engine.setLocal('attrs', element.attributes.get(element.ns, {}))
    if not hasattr(element, tal_handler.compiled_name):
        element = tal_compiler(element, context)
    return element

def tal_finish(element, context):
    try:
        del element.attributes[tal_ns]
        for dprefix, duri in element.nsdecls.items():
            if duri == tal_ns:
                del element.nsdecls[dprefix]
    except (AttributeError, KeyError):
        pass
    try:
        context.engine.endScope()
    except IndexError:
        pass
    if getattr(element, 'ns', None) == tal_ns:
        return element.children
    else:
        return element


tal_handler = CompiledInformationHandler('tal', tal_ns)
#tal_handler.add_handler('on_error', handle_on_error)
tal_handler.add_handler('define', handle_define)
tal_handler.add_handler('condition', handle_condition)
tal_handler.add_handler('repeat', handle_repeat)
tal_handler.add_handler('content', handle_content)
tal_handler.add_handler('replace', handle_replace)
tal_handler.add_handler('attributes', handle_attributes)
tal_handler.add_handler('omit-tag', handle_omit_tag)
tal_handler.preprocess = tal_start
tal_handler.postprocess = tal_finish


# TAL compiler

#def compile_simple(element, context, value):
#    return value

def compile_define(element, context, value):
    data = []
    for item in split_multiple(value):
        scope, expr = split_spec(item, 'local', 'global')

        name, expr = expr.split(None, 1)
        data.append((scope=='global', name, context.engine.compile(expr)))
    return data

def compile_condition(element, context, value):
    return context.engine.compile(value)

def compile_repeat(element, context, value):
    name, value = value.split(' ', 1)
    return name, context.engine.compile(value)

def compile_replace_or_content(element, context, value):
    mode, value = text_or_structure(value)
    return mode == 'structure', context.engine.compile(value)

def compile_attributes(element, context, value):
    r = []
    for item in split_multiple(value):
        spec, expr = item.split(' ', 1)
        if ':' in spec:
            ns, name = spec.split(':', 1)
        else:
            ns, name = element.ns, spec
        expr = context.engine.compile(expr)
        r.append((ns, name, expr))
    return r

def compile_omit_tag(element, context, value):
    if value:
        return context.engine.compile(value)

#def compile_tal_start(element, context):
#    return element

#def compile_tal_finish(element, context):
#    return element


tal_compiler = AttributeCompiler('tal', tal_ns)
#tal_compiler.add_handler('on_error', compile_on_error)
tal_compiler.add_handler('define', compile_define)
tal_compiler.add_handler('condition', compile_condition)
tal_compiler.add_handler('repeat', compile_repeat)
tal_compiler.add_handler('content', compile_replace_or_content)
tal_compiler.add_handler('replace', compile_replace_or_content)
tal_compiler.add_handler('attributes', compile_attributes)
tal_compiler.add_handler('omit-tag', compile_omit_tag)
#tal_compiler.preprocess = compile_tal_start
#tal_compiler.postprocess = compile_tal_finish


__all__ = ('tal_handler', 'tal_compiler')

if __name__ == '__main__':
    import sys, pax, pax.paxtransform
    tal_tr_engine = pax.paxtransform.Engine()
    tal_tr_engine.add_handler(tal_ns, tal_handler)
    import DummyEngine
    class DumbContext:
        def __init__(self, tr_engine):
            self.engine = DummyEngine.DummyEngine()
            self.tr_engine = tr_engine
    if len(sys.argv) > 1:
        srcfile = open(sys.argv[1])
    else:
        srcfile = sys.stdin
        sys.stderr.write('Reading source from stdin\n')
    source = pax.xml2pax(srcfile.read())
    print pax.XML(tal_tr_engine.transform(source, DumbContext(tal_tr_engine)))
